@@ -1,6 +1,6 @@
 //
-//  Copyright (C) 2001,2002 HorizonLive.com, Inc.  All Rights Reserved.
-//  Copyright (C) 2001,2002 Constantin Kaplinsky.  All Rights Reserved.
+//  Copyright (C) 2001-2004 HorizonLive.com, Inc.  All Rights Reserved.
+//  Copyright (C) 2001-2006 Constantin Kaplinsky.  All Rights Reserved.
 //  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
 //  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 //
@@ -32,17 +32,65 @@ import java.util.zip.*;
 
 class RfbProto {
 
-  final String versionMsg = "RFB 003.003\n";
-  final static int ConnFailed = 0, NoAuth = 1, VncAuth = 2;
-  final static int VncAuthOK = 0, VncAuthFailed = 1, VncAuthTooMany = 2;
+  final static String
+    versionMsg_3_3 = "RFB 003.003\n",
+    versionMsg_3_7 = "RFB 003.007\n";
 
-  final static int FramebufferUpdate = 0, SetColourMapEntries = 1, Bell = 2,
-    ServerCutText = 3;
+  // Vendor signatures: standard VNC/RealVNC, TridiaVNC, and TightVNC
+  final static String
+    StandardVendor  = "STDV",
+    TridiaVncVendor = "TRDV",
+    TightVncVendor  = "TGHT";
 
-  final int SetPixelFormat = 0, FixColourMapEntries = 1, SetEncodings = 2,
-    FramebufferUpdateRequest = 3, KeyboardEvent = 4, PointerEvent = 5,
-    ClientCutText = 6;
+  // Security types
+  final static int
+    SecTypeInvalid = 0,
+    SecTypeNone    = 1,
+    SecTypeVncAuth = 2,
+    SecTypeTight   = 16;
 
+  // Supported tunneling types
+  final static int
+    NoTunneling = 0;
+  final static String
+    SigNoTunneling = "NOTUNNEL";
+
+  // Supported authentication types
+  final static int
+    AuthNone      = 1,
+    AuthVNC       = 2,
+    AuthUnixLogin = 129;
+  final static String
+    SigAuthNone      = "NOAUTH__",
+    SigAuthVNC       = "VNCAUTH_",
+    SigAuthUnixLogin = "ULGNAUTH";
+
+  // VNC authentication results
+  final static int
+    VncAuthOK      = 0,
+    VncAuthFailed  = 1,
+    VncAuthTooMany = 2;
+
+  // Server-to-client messages
+  final static int
+    FramebufferUpdate   = 0,
+    SetColourMapEntries = 1,
+    Bell                = 2,
+    ServerCutText       = 3,
+    ClientAction        = 0x20;
+
+  // Client-to-server messages
+  final static int
+    SetPixelFormat           = 0,
+    FixColourMapEntries      = 1,
+    SetEncodings             = 2,
+    FramebufferUpdateRequest = 3,
+    KeyboardEvent            = 4,
+    PointerEvent             = 5,
+    ClientCutText            = 6,
+    RequestFBSize            = 0x20;
+
+  // Supported encodings and pseudo-encodings
   final static int
     EncodingRaw            = 0,
     EncodingCopyRect       = 1,
@@ -51,6 +99,7 @@ class RfbProto {
     EncodingHextile        = 5,
     EncodingZlib           = 6,
     EncodingTight          = 7,
+    EncodingZRLE           = 16,
     EncodingCompressLevel0 = 0xFFFFFF00,
     EncodingQualityLevel0  = 0xFFFFFFE0,
     EncodingXCursor        = 0xFFFFFF10,
@@ -58,24 +107,44 @@ class RfbProto {
     EncodingPointerPos     = 0xFFFFFF18,
     EncodingLastRect       = 0xFFFFFF20,
     EncodingNewFBSize      = 0xFFFFFF21;
+  final static String
+    SigEncodingRaw            = "RAW_____",
+    SigEncodingCopyRect       = "COPYRECT",
+    SigEncodingRRE            = "RRE_____",
+    SigEncodingCoRRE          = "CORRE___",
+    SigEncodingHextile        = "HEXTILE_",
+    SigEncodingZlib           = "ZLIB____",
+    SigEncodingTight          = "TIGHT___",
+    SigEncodingZRLE           = "ZRLE____",
+    SigEncodingCompressLevel0 = "COMPRLVL",
+    SigEncodingQualityLevel0  = "JPEGQLVL",
+    SigEncodingXCursor        = "X11CURSR",
+    SigEncodingRichCursor     = "RCHCURSR",
+    SigEncodingPointerPos     = "POINTPOS",
+    SigEncodingLastRect       = "LASTRECT",
+    SigEncodingNewFBSize      = "NEWFBSIZ";
 
-  final static int MaxNormalEncoding = 7;
+  final static int MaxNormalEncoding = 255;
 
-  final int HextileRaw			= (1 << 0);
-  final int HextileBackgroundSpecified	= (1 << 1);
-  final int HextileForegroundSpecified	= (1 << 2);
-  final int HextileAnySubrects		= (1 << 3);
-  final int HextileSubrectsColoured	= (1 << 4);
+  // Contstants used in the Hextile decoder
+  final static int
+    HextileRaw                 = 1,
+    HextileBackgroundSpecified = 2,
+    HextileForegroundSpecified = 4,
+    HextileAnySubrects         = 8,
+    HextileSubrectsColoured    = 16;
 
-  final static int TightExplicitFilter  = 0x04;
-  final static int TightFill            = 0x08;
-  final static int TightJpeg            = 0x09;
-  final static int TightMaxSubencoding  = 0x09;
-  final static int TightFilterCopy      = 0x00;
-  final static int TightFilterPalette   = 0x01;
-  final static int TightFilterGradient  = 0x02;
+  // Contstants used in the Tight decoder
+  final static int TightMinToCompress = 12;
+  final static int
+    TightExplicitFilter = 0x04,
+    TightFill           = 0x08,
+    TightJpeg           = 0x09,
+    TightMaxSubencoding = 0x09,
+    TightFilterCopy     = 0x00,
+    TightFilterPalette  = 0x01,
+    TightFilterGradient = 0x02;
 
-  final static int TightMinToCompress   = 12;
 
   String host;
   int port;
@@ -93,12 +162,12 @@ class RfbProto {
   boolean brokenKeyPressed = false;
 
   // This will be set to true on the first framebuffer update
-  // containing Zlib- or Tight-encoded data.
+  // containing Zlib-, ZRLE- or Tight-encoded data.
   boolean wereZlibUpdates = false;
 
   // This will be set to false if the startSession() was called after
-  // we have received at least one Zlib- or Tight-encoded framebuffer
-  // update.
+  // we have received at least one Zlib-, ZRLE- or Tight-encoded
+  // framebuffer update.
   boolean recordFromBeginning = true;
 
   // This fields are needed to show warnings about inefficiently saved
@@ -114,6 +183,19 @@ class RfbProto {
   // playback.
   int numUpdatesInSession;
 
+  // Measuring network throughput.
+  boolean timing;
+  long timeWaitedIn100us;
+  long timedKbits;
+
+  // Protocol version and TightVNC-specific protocol options.
+  int serverMajor, serverMinor;
+  int clientMajor, clientMinor;
+  boolean protocolTightVNC;
+  CapsContainer tunnelCaps, authCaps;
+  CapsContainer serverMsgCaps, clientMsgCaps;
+  CapsContainer encodingCaps;
+
   // If true, informs that the RFB socket was closed.
   private boolean closed;
 
@@ -126,7 +208,21 @@ class RfbProto {
     host = h;
     port = p;
 
-    if (viewer.socketFactory == null) {
+    if (! viewer.disableSSL) {
+	System.out.println("new SSLSocketToMe");
+	SSLSocketToMe ssl;
+	try {
+		ssl = new SSLSocketToMe(host, port, v);
+	} catch (Exception e) {
+		throw new IOException(e.getMessage());
+	}
+
+	try {
+		sock = ssl.connectSock();
+	} catch (Exception es) {
+		throw new IOException(es.getMessage());
+	}
+    } else if (viewer.socketFactory == null) {
       sock = new Socket(host, port);
     } else {
       try {
@@ -144,6 +240,10 @@ class RfbProto {
     is = new DataInputStream(new BufferedInputStream(sock.getInputStream(),
 						     16384));
     os = sock.getOutputStream();
+
+    timing = false;
+    timeWaitedIn100us = 5;
+    timedKbits = 0;
   }
 
 
@@ -169,13 +269,11 @@ class RfbProto {
   // Read server's protocol version message
   //
 
-  int serverMajor, serverMinor;
-
   void readVersionMsg() throws Exception {
 
     byte[] b = new byte[12];
 
-    is.readFully(b);
+    readFully(b);
 
     if ((b[0] != 'R') || (b[1] != 'F') || (b[2] != 'B') || (b[3] != ' ')
 	|| (b[4] < '0') || (b[4] > '9') || (b[5] < '0') || (b[5] > '9')
@@ -184,11 +282,15 @@ class RfbProto {
 	|| (b[10] < '0') || (b[10] > '9') || (b[11] != '\n'))
     {
       throw new Exception("Host " + host + " port " + port +
-			  " is not an RFB server");
+			  " is not an RFB server: " + b);
     }
 
     serverMajor = (b[4] - '0') * 100 + (b[5] - '0') * 10 + (b[6] - '0');
     serverMinor = (b[8] - '0') * 100 + (b[9] - '0') * 10 + (b[10] - '0');
+
+    if (serverMajor < 3) {
+      throw new Exception("RFB server does not support protocol version 3");
+    }
   }
 
 
@@ -197,36 +299,216 @@ class RfbProto {
   //
 
   void writeVersionMsg() throws IOException {
-    os.write(versionMsg.getBytes());
+    clientMajor = 3;
+    if (serverMajor > 3 || serverMinor >= 7) {
+      clientMinor = 7;
+      os.write(versionMsg_3_7.getBytes());
+    } else {
+      clientMinor = 3;
+      os.write(versionMsg_3_3.getBytes());
+    }
+    protocolTightVNC = false;
   }
 
 
   //
-  // Find out the authentication scheme.
+  // Negotiate the authentication scheme.
   //
 
-  int readAuthScheme() throws Exception {
-    int authScheme = is.readInt();
+  int negotiateSecurity() throws Exception {
+    return (clientMinor >= 7) ?
+      selectSecurityType() : readSecurityType();
+  }
 
-    switch (authScheme) {
+  //
+  // Read security type from the server (protocol version 3.3).
+  //
 
-    case ConnFailed:
-      int reasonLen = is.readInt();
-      byte[] reason = new byte[reasonLen];
-      is.readFully(reason);
-      throw new Exception(new String(reason));
-
-    case NoAuth:
-    case VncAuth:
-      return authScheme;
+  int readSecurityType() throws Exception {
+    int secType = is.readInt();
 
+    switch (secType) {
+    case SecTypeInvalid:
+      readConnFailedReason();
+      return SecTypeInvalid;	// should never be executed
+    case SecTypeNone:
+    case SecTypeVncAuth:
+      return secType;
     default:
-      throw new Exception("Unknown authentication scheme from RFB server: " +
-			  authScheme);
-
+      throw new Exception("Unknown security type from RFB server: " + secType);
     }
   }
 
+  //
+  // Select security type from the server's list (protocol version 3.7).
+  //
+
+  int selectSecurityType() throws Exception {
+    int secType = SecTypeInvalid;
+
+    // Read the list of secutiry types.
+    int nSecTypes = is.readUnsignedByte();
+    if (nSecTypes == 0) {
+      readConnFailedReason();
+      return SecTypeInvalid;	// should never be executed
+    }
+    byte[] secTypes = new byte[nSecTypes];
+    readFully(secTypes);
+
+    // Find out if the server supports TightVNC protocol extensions
+    for (int i = 0; i < nSecTypes; i++) {
+      if (secTypes[i] == SecTypeTight) {
+	protocolTightVNC = true;
+	os.write(SecTypeTight);
+	return SecTypeTight;
+      }
+    }
+
+    // Find first supported security type.
+    for (int i = 0; i < nSecTypes; i++) {
+      if (secTypes[i] == SecTypeNone || secTypes[i] == SecTypeVncAuth) {
+	secType = secTypes[i];
+	break;
+      }
+    }
+
+    if (secType == SecTypeInvalid) {
+      throw new Exception("Server did not offer supported security type");
+    } else {
+      os.write(secType);
+    }
+
+    return secType;
+  }
+
+  //
+  // Read the string describing the reason for a connection failure,
+  // and throw an exception.
+  //
+
+  void readConnFailedReason() throws Exception {
+    int reasonLen = is.readInt();
+    byte[] reason = new byte[reasonLen];
+    readFully(reason);
+    throw new Exception(new String(reason));
+  }
+
+  //
+  // Initialize capability lists (protocol 3.7t).
+  //
+
+  void initCapabilities() {
+    tunnelCaps    = new CapsContainer();
+    authCaps      = new CapsContainer();
+    serverMsgCaps = new CapsContainer();
+    clientMsgCaps = new CapsContainer();
+    encodingCaps  = new CapsContainer();
+
+    // Supported authentication methods
+    authCaps.add(AuthNone, StandardVendor, SigAuthNone,
+		 "No authentication");
+    authCaps.add(AuthVNC, StandardVendor, SigAuthVNC,
+		 "Standard VNC password authentication");
+    authCaps.add(AuthUnixLogin, TightVncVendor, SigAuthUnixLogin,
+		 "Login-style Unix authentication");
+
+    // Supported encoding types
+    encodingCaps.add(EncodingCopyRect, StandardVendor,
+		     SigEncodingCopyRect, "Standard CopyRect encoding");
+    encodingCaps.add(EncodingRRE, StandardVendor,
+		     SigEncodingRRE, "Standard RRE encoding");
+    encodingCaps.add(EncodingCoRRE, StandardVendor,
+		     SigEncodingCoRRE, "Standard CoRRE encoding");
+    encodingCaps.add(EncodingHextile, StandardVendor,
+		     SigEncodingHextile, "Standard Hextile encoding");
+    encodingCaps.add(EncodingZRLE, StandardVendor,
+		     SigEncodingZRLE, "Standard ZRLE encoding");
+    encodingCaps.add(EncodingZlib, TridiaVncVendor,
+		     SigEncodingZlib, "Zlib encoding");
+    encodingCaps.add(EncodingTight, TightVncVendor,
+		     SigEncodingTight, "Tight encoding");
+
+    // Supported pseudo-encoding types
+    encodingCaps.add(EncodingCompressLevel0, TightVncVendor,
+		     SigEncodingCompressLevel0, "Compression level");
+    encodingCaps.add(EncodingQualityLevel0, TightVncVendor,
+		     SigEncodingQualityLevel0, "JPEG quality level");
+    encodingCaps.add(EncodingXCursor, TightVncVendor,
+		     SigEncodingXCursor, "X-style cursor shape update");
+    encodingCaps.add(EncodingRichCursor, TightVncVendor,
+		     SigEncodingRichCursor, "Rich-color cursor shape update");
+    encodingCaps.add(EncodingPointerPos, TightVncVendor,
+		     SigEncodingPointerPos, "Pointer position update");
+    encodingCaps.add(EncodingLastRect, TightVncVendor,
+		     SigEncodingLastRect, "LastRect protocol extension");
+    encodingCaps.add(EncodingNewFBSize, TightVncVendor,
+		     SigEncodingNewFBSize, "Framebuffer size change");
+  }
+
+  //
+  // Setup tunneling (protocol version 3.7t)
+  //
+
+  void setupTunneling() throws IOException {
+    int nTunnelTypes = is.readInt();
+    if (nTunnelTypes != 0) {
+      readCapabilityList(tunnelCaps, nTunnelTypes);
+
+      // We don't support tunneling yet.
+      writeInt(NoTunneling);
+    }
+  }
+
+  //
+  // Negotiate authentication scheme (protocol version 3.7t)
+  //
+
+  int negotiateAuthenticationTight() throws Exception {
+    int nAuthTypes = is.readInt();
+    if (nAuthTypes == 0)
+      return AuthNone;
+
+    readCapabilityList(authCaps, nAuthTypes);
+    for (int i = 0; i < authCaps.numEnabled(); i++) {
+      int authType = authCaps.getByOrder(i);
+      if (authType == AuthNone ||
+	  authType == AuthVNC  ||
+	  authType == AuthUnixLogin) {
+	writeInt(authType);
+	return authType;
+      }
+    }
+    throw new Exception("No suitable authentication scheme found");
+  }
+
+  //
+  // Read a capability list (protocol version 3.7t)
+  //
+
+  void readCapabilityList(CapsContainer caps, int count) throws IOException {
+    int code;
+    byte[] vendor = new byte[4];
+    byte[] name = new byte[8];
+    for (int i = 0; i < count; i++) {
+      code = is.readInt();
+      readFully(vendor);
+      readFully(name);
+      caps.enable(new CapabilityInfo(code, vendor, name));
+    }
+  }
+
+  //
+  // Write a 32-bit integer into the output stream.
+  //
+
+  void writeInt(int value) throws IOException {
+    byte[] b = new byte[4];
+    b[0] = (byte) ((value >> 24) & 0xff);
+    b[1] = (byte) ((value >> 16) & 0xff);
+    b[2] = (byte) ((value >> 8) & 0xff);
+    b[3] = (byte) (value & 0xff);
+    os.write(b);
+  }
 
   //
   // Write the client initialisation message
@@ -266,12 +548,23 @@ class RfbProto {
     greenShift = is.readUnsignedByte();
     blueShift = is.readUnsignedByte();
     byte[] pad = new byte[3];
-    is.readFully(pad);
+    readFully(pad);
     int nameLength = is.readInt();
     byte[] name = new byte[nameLength];
-    is.readFully(name);
+    readFully(name);
     desktopName = new String(name);
 
+    // Read interaction capabilities (protocol 3.7t)
+    if (protocolTightVNC) {
+      int nServerMessageTypes = is.readUnsignedShort();
+      int nClientMessageTypes = is.readUnsignedShort();
+      int nEncodingTypes = is.readUnsignedShort();
+      is.readUnsignedShort();
+      readCapabilityList(serverMsgCaps, nServerMessageTypes);
+      readCapabilityList(clientMsgCaps, nClientMessageTypes);
+      readCapabilityList(encodingCaps, nEncodingTypes);
+    }
+
     inNormalProtocol = true;
   }
 
@@ -283,8 +576,8 @@ class RfbProto {
   void startSession(String fname) throws IOException {
     rec = new SessionRecorder(fname);
     rec.writeHeader();
-    rec.write(versionMsg.getBytes());
-    rec.writeIntBE(NoAuth);
+    rec.write(versionMsg_3_3.getBytes());
+    rec.writeIntBE(SecTypeNone);
     rec.writeShortBE(framebufferWidth);
     rec.writeShortBE(framebufferHeight);
     byte[] fbsServerInitMsg =	{
@@ -297,6 +590,11 @@ class RfbProto {
     rec.write(desktopName.getBytes());
     numUpdatesInSession = 0;
 
+    // FIXME: If there were e.g. ZRLE updates only, that should not
+    //        affect recording of Zlib and Tight updates. So, actually
+    //        we should maintain separate flags for Zlib, ZRLE and
+    //        Tight, instead of one ``wereZlibUpdates'' variable.
+    //
     if (wereZlibUpdates)
       recordFromBeginning = false;
 
@@ -378,6 +676,7 @@ class RfbProto {
     updateRectEncoding = is.readInt();
 
     if (updateRectEncoding == EncodingZlib ||
+        updateRectEncoding == EncodingZRLE ||
 	updateRectEncoding == EncodingTight)
       wereZlibUpdates = true;
 
@@ -442,10 +741,23 @@ class RfbProto {
 
   String readServerCutText() throws IOException {
     byte[] pad = new byte[3];
-    is.readFully(pad);
+    readFully(pad);
     int len = is.readInt();
     byte[] text = new byte[len];
-    is.readFully(text);
+    readFully(text);
+    return new String(text);
+  }
+
+  //
+  // Read a ClientAction message
+  //
+
+  String readClientAction() throws IOException {
+    byte[] pad = new byte[3];
+    readFully(pad);
+    int len = is.readInt();
+    byte[] text = new byte[len];
+    readFully(text);
     return new String(text);
   }
 
@@ -607,6 +919,22 @@ class RfbProto {
     os.write(b);
   }
 
+  //
+  // Write a RequestFBSize message
+  //
+
+  void writeRequestFBSize(int w, int h) throws IOException {
+    byte[] b = new byte[6];
+
+    b[0] = (byte) RequestFBSize;
+    b[1] = (byte) 42;
+    b[2] = (byte) ((w>>8) & 0x0ff);
+    b[3] = (byte) (w & 0x0ff);
+    b[4] = (byte) ((h>>8) & 0x0ff);
+    b[5] = (byte) (h & 0x0ff);
+
+    os.write(b);
+  }
 
   //
   // A buffer for putting pointer and keyboard events before being sent.  This
@@ -695,6 +1023,69 @@ class RfbProto {
     os.write(eventBuf, 0, eventBufLen);
   }
 
+  //
+  // Write a wheel event message.  We may need to send modifier key events
+  // around it to set the correct modifier state.
+  //
+
+  void writeWheelEvent(MouseWheelEvent evt) throws IOException {
+    int modifiers = evt.getModifiers();
+    int clicks = evt.getWheelRotation();
+    int mask;
+
+    int mask4 = 8;
+    int mask5 = 16;
+
+    if (clicks < 0) {
+      clicks = -clicks;
+      mask = mask4;
+    } else {
+      mask = mask5;
+    }
+
+    int x = evt.getX();
+    int y = evt.getY();
+
+    if (x < 0) x = 0;
+    if (y < 0) y = 0;
+
+    //
+    // Write the modifiers first.
+    //
+    eventBufLen = 0;
+    writeModifierKeyEvents(modifiers);
+    os.write(eventBuf, 0, eventBufLen);
+
+    for(int c=0; c<clicks; c++) {
+
+      eventBufLen = 0;
+
+      // Button Down
+      eventBuf[eventBufLen++] = (byte) PointerEvent;
+      eventBuf[eventBufLen++] = (byte) mask;
+      eventBuf[eventBufLen++] = (byte) ((x >> 8) & 0xff);
+      eventBuf[eventBufLen++] = (byte) (x & 0xff);
+      eventBuf[eventBufLen++] = (byte) ((y >> 8) & 0xff);
+      eventBuf[eventBufLen++] = (byte) (y & 0xff);
+
+      // Button Up
+      eventBuf[eventBufLen++] = (byte) PointerEvent;
+      eventBuf[eventBufLen++] = (byte) 0;
+      eventBuf[eventBufLen++] = (byte) ((x >> 8) & 0xff);
+      eventBuf[eventBufLen++] = (byte) (x & 0xff);
+      eventBuf[eventBufLen++] = (byte) ((y >> 8) & 0xff);
+      eventBuf[eventBufLen++] = (byte) (y & 0xff);
+
+      os.write(eventBuf, 0, eventBufLen);
+    }
+
+    //
+    // Always release all modifiers after an "up" event
+    //
+    eventBufLen = 0;
+    writeModifierKeyEvents(0);
+    os.write(eventBuf, 0, eventBufLen);
+  }
 
   //
   // Write a key event message.  We may need to send modifier key events
@@ -790,7 +1181,9 @@ class RfbProto {
 	// JDK1.1 on X incorrectly passes some keysyms straight through,
 	// so we do too.  JDK1.1.4 seems to have fixed this.
 	// The keysyms passed are 0xff00 .. XK_BackSpace .. XK_Delete
-	if ((key < 0xff00) || (key > 0xffff))
+	// Also, we pass through foreign currency keysyms (0x20a0..0x20af).
+	if ((key < 0xff00 || key > 0xffff) &&
+	    !(key >= 0x20a0 && key <= 0x20af))
 	  return;
       }
     }
@@ -911,5 +1304,56 @@ class RfbProto {
     }
     rec.write(buf, 0, bytes);
   }
-}
 
+  public void startTiming() {
+    timing = true;
+
+    // Carry over up to 1s worth of previous rate for smoothing.
+
+    if (timeWaitedIn100us > 10000) {
+      timedKbits = timedKbits * 10000 / timeWaitedIn100us;
+      timeWaitedIn100us = 10000;
+    }
+  }
+
+  public void stopTiming() {
+    timing = false; 
+    if (timeWaitedIn100us < timedKbits/2)
+      timeWaitedIn100us = timedKbits/2; // upper limit 20Mbit/s
+  }
+
+  public long kbitsPerSecond() {
+    return timedKbits * 10000 / timeWaitedIn100us;
+  }
+
+  public long timeWaited() {
+    return timeWaitedIn100us;
+  }
+
+  public void readFully(byte b[]) throws IOException {
+    readFully(b, 0, b.length);
+  }
+
+  public void readFully(byte b[], int off, int len) throws IOException {
+    long before = 0;
+    if (timing)
+      before = System.currentTimeMillis();
+
+    is.readFully(b, off, len);
+
+    if (timing) {
+      long after = System.currentTimeMillis();
+      long newTimeWaited = (after - before) * 10;
+      int newKbits = len * 8 / 1000;
+
+      // limit rate to between 10kbit/s and 40Mbit/s
+
+      if (newTimeWaited > newKbits*1000) newTimeWaited = newKbits*1000;
+      if (newTimeWaited < newKbits/4)    newTimeWaited = newKbits/4;
+
+      timeWaitedIn100us += newTimeWaited;
+      timedKbits += newKbits;
+    }
+  }
+
+}
