@@ -1,5 +1,6 @@
 //
-//  Copyright (C) 2001,2002 HorizonLive.com, Inc.  All Rights Reserved.
+//  Copyright (C) 2004 Horizon Wimba.  All Rights Reserved.
+//  Copyright (C) 2001-2003 HorizonLive.com, Inc.  All Rights Reserved.
 //  Copyright (C) 2001,2002 Constantin Kaplinsky.  All Rights Reserved.
 //  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
 //  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
@@ -27,13 +28,14 @@ import java.io.*;
 import java.lang.*;
 import java.util.zip.*;
 
+import java.util.Collections;
 
 //
 // VncCanvas is a subclass of Canvas which draws a VNC desktop on it.
 //
 
 class VncCanvas extends Canvas
-  implements KeyListener, MouseListener, MouseMotionListener {
+  implements KeyListener, MouseListener, MouseMotionListener, MouseWheelListener {
 
   VncViewer viewer;
   RfbProto rfb;
@@ -49,6 +51,14 @@ class VncCanvas extends Canvas
   byte[] pixels8;
   int[] pixels24;
 
+  // ZRLE encoder's data.
+  byte[] zrleBuf;
+  int zrleBufLen = 0;
+  byte[] zrleTilePixels8;
+  int[] zrleTilePixels24;
+  ZlibInStream zrleInStream;
+  boolean zrleRecWarningShown = false;
+
   // Zlib encoder's data.
   byte[] zlibBuf;
   int zlibBufLen = 0;
@@ -80,6 +90,21 @@ class VncCanvas extends Canvas
     cm8 = new DirectColorModel(8, 7, (7 << 3), (3 << 6));
     cm24 = new DirectColorModel(24, 0xFF0000, 0x00FF00, 0x0000FF);
 
+    // kludge to not show any Java cursor in the canvas since we are
+    // showing the soft cursor (should be a user setting...)
+    Cursor dot = Toolkit.getDefaultToolkit().createCustomCursor(
+        Toolkit.getDefaultToolkit().createImage(new byte[4]), new Point(0,0),
+        "dot");
+    this.setCursor(dot);
+
+    // while we are at it... get rid of the keyboard traversals that
+    // make it so we can't type a Tab character:
+    this.setFocusTraversalKeysEnabled(false);
+    this.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
+        Collections.EMPTY_SET);
+    this.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
+        Collections.EMPTY_SET);
+
     colors = new Color[256];
     for (int i = 0; i < 256; i++)
       colors[i] = new Color(cm8.getRGB(i));
@@ -168,9 +193,11 @@ class VncCanvas extends Canvas
       inputEnabled = true;
       addMouseListener(this);
       addMouseMotionListener(this);
+      addMouseWheelListener(this);
       if (viewer.showControls) {
 	viewer.buttonPanel.enableRemoteAccessControls(true);
       }
+      createSoftCursor();	// scaled cursor
     } else if (!enable && inputEnabled) {
       inputEnabled = false;
       removeMouseListener(this);
@@ -178,6 +205,7 @@ class VncCanvas extends Canvas
       if (viewer.showControls) {
 	viewer.buttonPanel.enableRemoteAccessControls(false);
       }
+      createSoftCursor();	// non-scaled cursor
     }
   }
 
@@ -192,8 +220,7 @@ class VncCanvas extends Canvas
     updateFramebufferSize();
   }
 
-  void updateFramebufferSize() {
-
+  void unsafeUpdateFramebufferSize() {
     // Useful shortcuts.
     int fbWidth = rfb.framebufferWidth;
     int fbHeight = rfb.framebufferHeight;
@@ -202,12 +229,12 @@ class VncCanvas extends Canvas
     // its geometry should be changed. It's not necessary to replace
     // existing image if only pixel format should be changed.
     if (memImage == null) {
-      memImage = viewer.createImage(fbWidth, fbHeight);
+      memImage = viewer.vncContainer.createImage(fbWidth, fbHeight);
       memGraphics = memImage.getGraphics();
     } else if (memImage.getWidth(null) != fbWidth ||
 	       memImage.getHeight(null) != fbHeight) {
       synchronized(memImage) {
-	memImage = viewer.createImage(fbWidth, fbHeight);
+	memImage = viewer.vncContainer.createImage(fbWidth, fbHeight);
 	memGraphics = memImage.getGraphics();
       }
     }
@@ -220,15 +247,19 @@ class VncCanvas extends Canvas
 
       pixelsSource =
 	new MemoryImageSource(fbWidth, fbHeight, cm8, pixels8, 0, fbWidth);
+      zrleTilePixels24 = null;
+      zrleTilePixels8 = new byte[64 * 64];
     } else {
       pixels8 = null;
       pixels24 = new int[fbWidth * fbHeight];
 
       pixelsSource =
 	new MemoryImageSource(fbWidth, fbHeight, cm24, pixels24, 0, fbWidth);
+      zrleTilePixels8 = null;
+      zrleTilePixels24 = new int[64 * 64];
     }
     pixelsSource.setAnimated(true);
-    rawPixelsImage = createImage(pixelsSource);
+    rawPixelsImage = Toolkit.getDefaultToolkit().createImage(pixelsSource);
 
     // Update the size of desktop containers.
     if (viewer.inSeparateFrame) {
@@ -237,9 +268,56 @@ class VncCanvas extends Canvas
     } else {
       setSize(fbWidth, fbHeight);
     }
+    viewer.forceSize();
     viewer.moveFocusToDesktop();
+
+    return;
   }
 
+
+  void updateFramebufferSize() {
+    int retry;
+
+    // Try three times to reallocate the framebuffer.
+    // Do garbage collection first.
+    //
+    for (retry=0; retry < 3; retry++) {
+      System.gc();
+      try {
+        unsafeUpdateFramebufferSize();
+        return;
+      }
+
+      // If we can't allocate memory for the new framebuffer, then
+      // try deleting all of the existing framebuffer memory first.
+      //
+      catch(java.lang.OutOfMemoryError e) {
+        System.out.println("WARNING: Out of memory.  Retrying.\n");
+        //synchronized(memImage) {
+          memImage = viewer.vncContainer.createImage(1, 1);
+          memGraphics = memImage.getGraphics();
+        //}
+        pixels8  = new byte[1];
+        pixels24 = new int[1];
+        if (bytesPixel == 1) {
+          pixelsSource = new MemoryImageSource(1, 1, cm8, pixels8, 0, 1);
+        } else {
+          pixelsSource = new MemoryImageSource(1, 1, cm24, pixels24, 0, 1);
+        }
+        pixelsSource.setAnimated(true);
+        rawPixelsImage = Toolkit.getDefaultToolkit().createImage(pixelsSource);
+        System.gc();
+      }
+    }
+    System.out.println("ERROR: Screen size too big.  We're out of memory.\n");
+    int w = rfb.framebufferWidth * 8 / 10;
+    int h = rfb.framebufferHeight * 8 / 10;
+    System.out.println("Requesting resize to " +
+                       Integer.toString(w) + "x" + Integer.toString(h));
+    viewer.requestResize(w,h);
+  }
+
+
   void resizeDesktopFrame() {
     setSize(rfb.framebufferWidth, rfb.framebufferHeight);
 
@@ -323,6 +401,17 @@ class VncCanvas extends Canvas
 	    break;
 
 	  if (rfb.updateRectEncoding == rfb.EncodingNewFBSize) {
+
+            if (viewer.setServerSize) {
+              if (rw == rfb.framebufferWidth && rh == rfb.framebufferHeight) {
+                System.out.println("SetServerSize: New size is right.");
+                viewer.setServerSize = false;
+              } else {
+                System.out.println("SetServerSize: Requesting server resize.");
+                viewer.requestResize(rfb.framebufferWidth,rfb.framebufferHeight);
+                break;
+              }
+            }
 	    rfb.setFramebufferSize(rw, rh);
 	    updateFramebufferSize();
 	    break;
@@ -340,6 +429,8 @@ class VncCanvas extends Canvas
 	    continue;
 	  }
 
+          rfb.startTiming();
+
 	  switch (rfb.updateRectEncoding) {
 	  case RfbProto.EncodingRaw:
 	    handleRawRect(rx, ry, rw, rh);
@@ -356,6 +447,9 @@ class VncCanvas extends Canvas
 	  case RfbProto.EncodingHextile:
 	    handleHextileRect(rx, ry, rw, rh);
 	    break;
+	  case RfbProto.EncodingZRLE:
+	    handleZRLERect(rx, ry, rw, rh);
+	    break;
 	  case RfbProto.EncodingZlib:
             handleZlibRect(rx, ry, rw, rh);
 	    break;
@@ -366,6 +460,8 @@ class VncCanvas extends Canvas
 	    throw new Exception("Unknown RFB rectangle encoding " +
 				rfb.updateRectEncoding);
 	  }
+
+          rfb.stopTiming();
 	}
 
 	boolean fullUpdateNeeded = false;
@@ -397,6 +493,8 @@ class VncCanvas extends Canvas
 	  fullUpdateNeeded = true;
 	}
 
+        viewer.autoSelectEncodings();
+
 	rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
 					  rfb.framebufferHeight,
 					  !fullUpdateNeeded);
@@ -415,6 +513,11 @@ class VncCanvas extends Canvas
 	viewer.clipboard.setCutText(s);
 	break;
 
+      case RfbProto.ClientAction:
+	String action = rfb.readClientAction();
+	viewer.clientAction(action);
+	break;
+
       default:
 	throw new Exception("Unknown RFB message type " + msgType);
       }
@@ -436,7 +539,7 @@ class VncCanvas extends Canvas
 
     if (bytesPixel == 1) {
       for (int dy = y; dy < y + h; dy++) {
-	rfb.is.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
+	rfb.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
 	if (rfb.rec != null) {
 	  rfb.rec.write(pixels8, dy * rfb.framebufferWidth + x, w);
 	}
@@ -445,7 +548,7 @@ class VncCanvas extends Canvas
       byte[] buf = new byte[w * 4];
       int i, offset;
       for (int dy = y; dy < y + h; dy++) {
-	rfb.is.readFully(buf);
+	rfb.readFully(buf);
 	if (rfb.rec != null) {
 	  rfb.rec.write(buf);
 	}
@@ -486,7 +589,7 @@ class VncCanvas extends Canvas
     int nSubrects = rfb.is.readInt();
 
     byte[] bg_buf = new byte[bytesPixel];
-    rfb.is.readFully(bg_buf);
+    rfb.readFully(bg_buf);
     Color pixel;
     if (bytesPixel == 1) {
       pixel = colors[bg_buf[0] & 0xFF];
@@ -497,7 +600,7 @@ class VncCanvas extends Canvas
     memGraphics.fillRect(x, y, w, h);
 
     byte[] buf = new byte[nSubrects * (bytesPixel + 8)];
-    rfb.is.readFully(buf);
+    rfb.readFully(buf);
     DataInputStream ds = new DataInputStream(new ByteArrayInputStream(buf));
 
     if (rfb.rec != null) {
@@ -537,7 +640,7 @@ class VncCanvas extends Canvas
     int nSubrects = rfb.is.readInt();
 
     byte[] bg_buf = new byte[bytesPixel];
-    rfb.is.readFully(bg_buf);
+    rfb.readFully(bg_buf);
     Color pixel;
     if (bytesPixel == 1) {
       pixel = colors[bg_buf[0] & 0xFF];
@@ -548,7 +651,7 @@ class VncCanvas extends Canvas
     memGraphics.fillRect(x, y, w, h);
 
     byte[] buf = new byte[nSubrects * (bytesPixel + 4)];
-    rfb.is.readFully(buf);
+    rfb.readFully(buf);
 
     if (rfb.rec != null) {
       rfb.rec.writeIntBE(nSubrects);
@@ -629,7 +732,7 @@ class VncCanvas extends Canvas
     // Read and draw the background if specified.
     byte[] cbuf = new byte[bytesPixel];
     if ((subencoding & rfb.HextileBackgroundSpecified) != 0) {
-      rfb.is.readFully(cbuf);
+      rfb.readFully(cbuf);
       if (bytesPixel == 1) {
 	hextile_bg = colors[cbuf[0] & 0xFF];
       } else {
@@ -644,7 +747,7 @@ class VncCanvas extends Canvas
 
     // Read the foreground color if specified.
     if ((subencoding & rfb.HextileForegroundSpecified) != 0) {
-      rfb.is.readFully(cbuf);
+      rfb.readFully(cbuf);
       if (bytesPixel == 1) {
 	hextile_fg = colors[cbuf[0] & 0xFF];
       } else {
@@ -665,7 +768,7 @@ class VncCanvas extends Canvas
       bufsize += nSubrects * bytesPixel;
     }
     byte[] buf = new byte[bufsize];
-    rfb.is.readFully(buf);
+    rfb.readFully(buf);
     if (rfb.rec != null) {
       rfb.rec.writeByte(nSubrects);
       rfb.rec.write(buf);
@@ -723,6 +826,237 @@ class VncCanvas extends Canvas
     }
   }
 
+  //
+  // Handle a ZRLE-encoded rectangle.
+  //
+  // FIXME: Currently, session recording is not fully supported for ZRLE.
+  //
+
+  void handleZRLERect(int x, int y, int w, int h) throws Exception {
+
+    if (zrleInStream == null)
+      zrleInStream = new ZlibInStream();
+
+    int nBytes = rfb.is.readInt();
+    if (nBytes > 64 * 1024 * 1024)
+      throw new Exception("ZRLE decoder: illegal compressed data size");
+
+    if (zrleBuf == null || zrleBufLen < nBytes) {
+      zrleBufLen = nBytes + 4096;
+      zrleBuf = new byte[zrleBufLen];
+    }
+
+    // FIXME: Do not wait for all the data before decompression.
+    rfb.readFully(zrleBuf, 0, nBytes);
+
+    if (rfb.rec != null) {
+      if (rfb.recordFromBeginning) {
+        rfb.rec.writeIntBE(nBytes);
+        rfb.rec.write(zrleBuf, 0, nBytes);
+      } else if (!zrleRecWarningShown) {
+        System.out.println("Warning: ZRLE session can be recorded" +
+                           " only from the beginning");
+        System.out.println("Warning: Recorded file may be corrupted");
+        zrleRecWarningShown = true;
+      }
+    }
+
+    zrleInStream.setUnderlying(new MemInStream(zrleBuf, 0, nBytes), nBytes);
+
+    for (int ty = y; ty < y+h; ty += 64) {
+
+      int th = Math.min(y+h-ty, 64);
+
+      for (int tx = x; tx < x+w; tx += 64) {
+
+        int tw = Math.min(x+w-tx, 64);
+
+        int mode = zrleInStream.readU8();
+        boolean rle = (mode & 128) != 0;
+        int palSize = mode & 127;
+        int[] palette = new int[128];
+
+        readZrlePalette(palette, palSize);
+
+        if (palSize == 1) {
+          int pix = palette[0];
+          Color c = (bytesPixel == 1) ?
+            colors[pix] : new Color(0xFF000000 | pix);
+          memGraphics.setColor(c);
+          memGraphics.fillRect(tx, ty, tw, th);
+          continue;
+        }
+
+        if (!rle) {
+          if (palSize == 0) {
+            readZrleRawPixels(tw, th);
+          } else {
+            readZrlePackedPixels(tw, th, palette, palSize);
+          }
+        } else {
+          if (palSize == 0) {
+            readZrlePlainRLEPixels(tw, th);
+          } else {
+            readZrlePackedRLEPixels(tw, th, palette);
+          }
+        }
+        handleUpdatedZrleTile(tx, ty, tw, th);
+      }
+    }
+
+    zrleInStream.reset();
+
+    scheduleRepaint(x, y, w, h);
+  }
+
+  int readPixel(InStream is) throws Exception {
+    int pix;
+    if (bytesPixel == 1) {
+      pix = is.readU8();
+    } else {
+      int p1 = is.readU8();
+      int p2 = is.readU8();
+      int p3 = is.readU8();
+      pix = (p3 & 0xFF) << 16 | (p2 & 0xFF) << 8 | (p1 & 0xFF);
+    }
+    return pix;
+  }
+
+  void readPixels(InStream is, int[] dst, int count) throws Exception {
+    int pix;
+    if (bytesPixel == 1) {
+      byte[] buf = new byte[count];
+      is.readBytes(buf, 0, count);
+      for (int i = 0; i < count; i++) {
+        dst[i] = (int)buf[i] & 0xFF;
+      }
+    } else {
+      byte[] buf = new byte[count * 3];
+      is.readBytes(buf, 0, count * 3);
+      for (int i = 0; i < count; i++) {
+        dst[i] = ((buf[i*3+2] & 0xFF) << 16 |
+                  (buf[i*3+1] & 0xFF) << 8 |
+                  (buf[i*3] & 0xFF));
+      }
+    }
+  }
+
+  void readZrlePalette(int[] palette, int palSize) throws Exception {
+    readPixels(zrleInStream, palette, palSize);
+  }
+
+  void readZrleRawPixels(int tw, int th) throws Exception {
+    if (bytesPixel == 1) {
+      zrleInStream.readBytes(zrleTilePixels8, 0, tw * th);
+    } else {
+      readPixels(zrleInStream, zrleTilePixels24, tw * th); ///
+    }
+  }
+
+  void readZrlePackedPixels(int tw, int th, int[] palette, int palSize)
+    throws Exception {
+
+    int bppp = ((palSize > 16) ? 8 :
+                ((palSize > 4) ? 4 : ((palSize > 2) ? 2 : 1)));
+    int ptr = 0;
+
+    for (int i = 0; i < th; i++) {
+      int eol = ptr + tw;
+      int b = 0;
+      int nbits = 0;
+
+      while (ptr < eol) {
+        if (nbits == 0) {
+          b = zrleInStream.readU8();
+          nbits = 8;
+        }
+        nbits -= bppp;
+        int index = (b >> nbits) & ((1 << bppp) - 1) & 127;
+        if (bytesPixel == 1) {
+          zrleTilePixels8[ptr++] = (byte)palette[index];
+        } else {
+          zrleTilePixels24[ptr++] = palette[index];
+        }
+      }
+    }
+  }
+
+  void readZrlePlainRLEPixels(int tw, int th) throws Exception {
+    int ptr = 0;
+    int end = ptr + tw * th;
+    while (ptr < end) {
+      int pix = readPixel(zrleInStream);
+      int len = 1;
+      int b;
+      do {
+        b = zrleInStream.readU8();
+        len += b;
+      } while (b == 255);
+
+      if (!(len <= end - ptr))
+        throw new Exception("ZRLE decoder: assertion failed" +
+                            " (len <= end-ptr)");
+
+      if (bytesPixel == 1) {
+        while (len-- > 0) zrleTilePixels8[ptr++] = (byte)pix;
+      } else {
+        while (len-- > 0) zrleTilePixels24[ptr++] = pix;
+      }
+    }
+  }
+
+  void readZrlePackedRLEPixels(int tw, int th, int[] palette)
+    throws Exception {
+
+    int ptr = 0;
+    int end = ptr + tw * th;
+    while (ptr < end) {
+      int index = zrleInStream.readU8();
+      int len = 1;
+      if ((index & 128) != 0) {
+        int b;
+        do {
+          b = zrleInStream.readU8();
+          len += b;
+        } while (b == 255);
+        
+        if (!(len <= end - ptr))
+          throw new Exception("ZRLE decoder: assertion failed" +
+                              " (len <= end - ptr)");
+      }
+
+      index &= 127;
+      int pix = palette[index];
+
+      if (bytesPixel == 1) {
+        while (len-- > 0) zrleTilePixels8[ptr++] = (byte)pix;
+      } else {
+        while (len-- > 0) zrleTilePixels24[ptr++] = pix;
+      }
+    }
+  }
+
+  //
+  // Copy pixels from zrleTilePixels8 or zrleTilePixels24, then update.
+  //
+
+  void handleUpdatedZrleTile(int x, int y, int w, int h) {
+    Object src, dst;
+    if (bytesPixel == 1) {
+      src = zrleTilePixels8; dst = pixels8;
+    } else {
+      src = zrleTilePixels24; dst = pixels24;
+    }
+    int offsetSrc = 0;
+    int offsetDst = (y * rfb.framebufferWidth + x);
+    for (int j = 0; j < h; j++) {
+      System.arraycopy(src, offsetSrc, dst, offsetDst, w);
+      offsetSrc += w;
+      offsetDst += rfb.framebufferWidth;
+    }
+    handleUpdatedPixels(x, y, w, h);
+  }
+
   //
   // Handle a Zlib-encoded rectangle.
   //
@@ -736,7 +1070,7 @@ class VncCanvas extends Canvas
       zlibBuf = new byte[zlibBufLen];
     }
 
-    rfb.is.readFully(zlibBuf, 0, nBytes);
+    rfb.readFully(zlibBuf, 0, nBytes);
 
     if (rfb.rec != null && rfb.recordFromBeginning) {
       rfb.rec.writeIntBE(nBytes);
@@ -818,7 +1152,7 @@ class VncCanvas extends Canvas
 	}
       } else {
 	byte[] buf = new byte[3];
-	rfb.is.readFully(buf);
+	rfb.readFully(buf);
 	if (rfb.rec != null) {
 	  rfb.rec.write(buf);
 	}
@@ -836,7 +1170,7 @@ class VncCanvas extends Canvas
 
       // Read JPEG data.
       byte[] jpegData = new byte[rfb.readCompactLen()];
-      rfb.is.readFully(jpegData);
+      rfb.readFully(jpegData);
       if (rfb.rec != null) {
 	if (!rfb.recordFromBeginning) {
 	  rfb.recordCompactLen(jpegData.length);
@@ -887,13 +1221,13 @@ class VncCanvas extends Canvas
 	  if (numColors != 2) {
 	    throw new Exception("Incorrect tight palette size: " + numColors);
 	  }
-	  rfb.is.readFully(palette8);
+	  rfb.readFully(palette8);
 	  if (rfb.rec != null) {
 	    rfb.rec.write(palette8);
 	  }
 	} else {
 	  byte[] buf = new byte[numColors * 3];
-	  rfb.is.readFully(buf);
+	  rfb.readFully(buf);
 	  if (rfb.rec != null) {
 	    rfb.rec.write(buf);
 	  }
@@ -921,7 +1255,7 @@ class VncCanvas extends Canvas
       if (numColors != 0) {
 	// Indexed colors.
 	byte[] indexedData = new byte[dataSize];
-	rfb.is.readFully(indexedData);
+	rfb.readFully(indexedData);
 	if (rfb.rec != null) {
 	  rfb.rec.write(indexedData);
 	}
@@ -945,7 +1279,7 @@ class VncCanvas extends Canvas
       } else if (useGradient) {
 	// "Gradient"-processed data
 	byte[] buf = new byte[w * h * 3];
-	rfb.is.readFully(buf);
+	rfb.readFully(buf);
 	if (rfb.rec != null) {
 	  rfb.rec.write(buf);
 	}
@@ -954,7 +1288,7 @@ class VncCanvas extends Canvas
 	// Raw truecolor data.
 	if (bytesPixel == 1) {
 	  for (int dy = y; dy < y + h; dy++) {
-	    rfb.is.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
+	    rfb.readFully(pixels8, dy * rfb.framebufferWidth + x, w);
 	    if (rfb.rec != null) {
 	      rfb.rec.write(pixels8, dy * rfb.framebufferWidth + x, w);
 	    }
@@ -963,7 +1297,7 @@ class VncCanvas extends Canvas
 	  byte[] buf = new byte[w * 3];
 	  int i, offset;
 	  for (int dy = y; dy < y + h; dy++) {
-	    rfb.is.readFully(buf);
+	    rfb.readFully(buf);
 	    if (rfb.rec != null) {
 	      rfb.rec.write(buf);
 	    }
@@ -981,7 +1315,7 @@ class VncCanvas extends Canvas
       // Data was compressed with zlib.
       int zlibDataLen = rfb.readCompactLen();
       byte[] zlibData = new byte[zlibDataLen];
-      rfb.is.readFully(zlibData);
+      rfb.readFully(zlibData);
       if (rfb.rec != null && rfb.recordFromBeginning) {
 	rfb.rec.write(zlibData);
       }
@@ -1187,6 +1521,9 @@ class VncCanvas extends Canvas
   public void mouseDragged(MouseEvent evt) {
     processLocalMouseEvent(evt, true);
   }
+  public void mouseWheelMoved(MouseWheelEvent evt) {
+    processLocalWheelEvent(evt, true);
+  }
 
   public void processLocalKeyEvent(KeyEvent evt) {
     if (viewer.rfb != null && rfb.inNormalProtocol) {
@@ -1234,6 +1571,22 @@ class VncCanvas extends Canvas
     }
   }
 
+  public void processLocalWheelEvent(MouseWheelEvent evt, boolean moved) {
+    if (viewer.rfb != null && rfb.inNormalProtocol) {
+      if (moved) {
+	softCursorMove(evt.getX(), evt.getY());
+      }
+      synchronized(rfb) {
+	try {
+	  rfb.writeWheelEvent(evt);
+	} catch (Exception e) {
+	  e.printStackTrace();
+	}
+	rfb.notify();
+      }
+    }
+  }
+
   //
   // Ignored events.
   //
@@ -1250,13 +1603,14 @@ class VncCanvas extends Canvas
 
   boolean showSoftCursor = false;
 
-  int[] softCursorPixels;
   MemoryImageSource softCursorSource;
   Image softCursor;
 
   int cursorX = 0, cursorY = 0;
   int cursorWidth, cursorHeight;
+  int origCursorWidth, origCursorHeight;
   int hotX, hotY;
+  int origHotX, origHotY;
 
   //
   // Handle cursor shape update (XCursor and RichCursor encodings).
@@ -1267,17 +1621,16 @@ class VncCanvas extends Canvas
 			    int xhot, int yhot, int width, int height)
     throws IOException {
 
-    int bytesPerRow = (width + 7) / 8;
-    int bytesMaskData = bytesPerRow * height;
-
     softCursorFree();
 
     if (width * height == 0)
       return;
 
     // Ignore cursor shape data if requested by user.
-
     if (viewer.options.ignoreCursorUpdates) {
+      int bytesPerRow = (width + 7) / 8;
+      int bytesMaskData = bytesPerRow * height;
+
       if (encodingType == rfb.EncodingXCursor) {
 	rfb.is.skipBytes(6 + bytesMaskData * 2);
       } else {
@@ -1288,14 +1641,42 @@ class VncCanvas extends Canvas
     }
 
     // Decode cursor pixel data.
+    softCursorSource = decodeCursorShape(encodingType, width, height);
 
-    softCursorPixels = new int[width * height];
+    // Set original (non-scaled) cursor dimensions.
+    origCursorWidth = width;
+    origCursorHeight = height;
+    origHotX = xhot;
+    origHotY = yhot;
+
+    // Create off-screen cursor image.
+    createSoftCursor();
+
+    // Show the cursor.
+    showSoftCursor = true;
+    repaint(viewer.deferCursorUpdates,
+	    cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
+  }
+
+  //
+  // decodeCursorShape(). Decode cursor pixel data and return
+  // corresponding MemoryImageSource instance.
+  //
+
+  synchronized MemoryImageSource
+    decodeCursorShape(int encodingType, int width, int height)
+    throws IOException {
+
+    int bytesPerRow = (width + 7) / 8;
+    int bytesMaskData = bytesPerRow * height;
+
+    int[] softCursorPixels = new int[width * height];
 
     if (encodingType == rfb.EncodingXCursor) {
 
       // Read foreground and background colors of the cursor.
       byte[] rgb = new byte[6];
-      rfb.is.readFully(rgb);
+      rfb.readFully(rgb);
       int[] colors = { (0xFF000000 | (rgb[3] & 0xFF) << 16 |
 			(rgb[4] & 0xFF) << 8 | (rgb[5] & 0xFF)),
 		       (0xFF000000 | (rgb[0] & 0xFF) << 16 |
@@ -1303,9 +1684,9 @@ class VncCanvas extends Canvas
 
       // Read pixel and mask data.
       byte[] pixBuf = new byte[bytesMaskData];
-      rfb.is.readFully(pixBuf);
+      rfb.readFully(pixBuf);
       byte[] maskBuf = new byte[bytesMaskData];
-      rfb.is.readFully(maskBuf);
+      rfb.readFully(maskBuf);
 
       // Decode pixel data into softCursorPixels[].
       byte pixByte, maskByte;
@@ -1339,9 +1720,9 @@ class VncCanvas extends Canvas
 
       // Read pixel and mask data.
       byte[] pixBuf = new byte[width * height * bytesPixel];
-      rfb.is.readFully(pixBuf);
+      rfb.readFully(pixBuf);
       byte[] maskBuf = new byte[bytesMaskData];
-      rfb.is.readFully(maskBuf);
+      rfb.readFully(maskBuf);
 
       // Decode pixel data into softCursorPixels[].
       byte pixByte, maskByte;
@@ -1356,9 +1737,9 @@ class VncCanvas extends Canvas
 		result = cm8.getRGB(pixBuf[i]);
 	      } else {
 		result = 0xFF000000 |
-		  (pixBuf[i * 4 + 1] & 0xFF) << 16 |
-		  (pixBuf[i * 4 + 2] & 0xFF) << 8 |
-		  (pixBuf[i * 4 + 3] & 0xFF);
+		  (pixBuf[i * 4 + 2] & 0xFF) << 16 |
+		  (pixBuf[i * 4 + 1] & 0xFF) << 8 |
+		  (pixBuf[i * 4 + 0] & 0xFF);
 	      }
 	    } else {
 	      result = 0;	// Transparent pixel
@@ -1372,9 +1753,9 @@ class VncCanvas extends Canvas
 	      result = cm8.getRGB(pixBuf[i]);
 	    } else {
 	      result = 0xFF000000 |
-		(pixBuf[i * 4 + 1] & 0xFF) << 16 |
-		(pixBuf[i * 4 + 2] & 0xFF) << 8 |
-		(pixBuf[i * 4 + 3] & 0xFF);
+		(pixBuf[i * 4 + 2] & 0xFF) << 16 |
+		(pixBuf[i * 4 + 1] & 0xFF) << 8 |
+		(pixBuf[i * 4 + 0] & 0xFF);
 	    }
 	  } else {
 	    result = 0;		// Transparent pixel
@@ -1385,25 +1766,50 @@ class VncCanvas extends Canvas
 
     }
 
-    // Draw the cursor on an off-screen image.
+    return new MemoryImageSource(width, height, softCursorPixels, 0, width);
+  }
 
-    softCursorSource =
-      new MemoryImageSource(width, height, softCursorPixels, 0, width);
-    softCursor = createImage(softCursorSource);
+  //
+  // createSoftCursor(). Assign softCursor new Image (scaled if necessary).
+  // Uses softCursorSource as a source for new cursor image.
+  //
 
-    // Set remaining data associated with cursor.
+  synchronized void
+    createSoftCursor() {
 
-    cursorWidth = width;
-    cursorHeight = height;
-    hotX = xhot;
-    hotY = yhot;
+    if (softCursorSource == null)
+      return;
 
-    showSoftCursor = true;
+    int scaleCursor = viewer.options.scaleCursor;
+    if (scaleCursor == 0 || !inputEnabled)
+      scaleCursor = 100;
 
-    // Show the cursor.
+    // Save original cursor coordinates.
+    int x = cursorX - hotX;
+    int y = cursorY - hotY;
+    int w = cursorWidth;
+    int h = cursorHeight;
 
-    repaint(viewer.deferCursorUpdates,
-	    cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
+    cursorWidth = (origCursorWidth * scaleCursor + 50) / 100;
+    cursorHeight = (origCursorHeight * scaleCursor + 50) / 100;
+    hotX = (origHotX * scaleCursor + 50) / 100;
+    hotY = (origHotY * scaleCursor + 50) / 100;
+    softCursor = Toolkit.getDefaultToolkit().createImage(softCursorSource);
+
+    if (scaleCursor != 100) {
+      softCursor = softCursor.getScaledInstance(cursorWidth, cursorHeight,
+						Image.SCALE_SMOOTH);
+    }
+
+    if (showSoftCursor) {
+      // Compute screen area to update.
+      x = Math.min(x, cursorX - hotX);
+      y = Math.min(y, cursorY - hotY);
+      w = Math.max(w, cursorWidth);
+      h = Math.max(h, cursorHeight);
+
+      repaint(viewer.deferCursorUpdates, x, y, w, h);
+    }
   }
 
   //
@@ -1411,15 +1817,16 @@ class VncCanvas extends Canvas
   //
 
   synchronized void softCursorMove(int x, int y) {
-    if (showSoftCursor) {
-      repaint(viewer.deferCursorUpdates,
-	      cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
-      repaint(viewer.deferCursorUpdates,
-	      x - hotX, y - hotY, cursorWidth, cursorHeight);
-    }
-
+    int oldX = cursorX;
+    int oldY = cursorY;
     cursorX = x;
     cursorY = y;
+    if (showSoftCursor) {
+      repaint(viewer.deferCursorUpdates,
+	      oldX - hotX, oldY - hotY, cursorWidth, cursorHeight);
+      repaint(viewer.deferCursorUpdates,
+	      cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
+    }
   }
 
   //
@@ -1431,7 +1838,6 @@ class VncCanvas extends Canvas
       showSoftCursor = false;
       softCursor = null;
       softCursorSource = null;
-      softCursorPixels = null;
 
       repaint(viewer.deferCursorUpdates,
 	      cursorX - hotX, cursorY - hotY, cursorWidth, cursorHeight);
