@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2001,2002 HorizonLive.com, Inc.  All Rights Reserved.
+//  Copyright (C) 2001-2004 HorizonLive.com, Inc.  All Rights Reserved.
 //  Copyright (C) 2002 Constantin Kaplinsky.  All Rights Reserved.
 //  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 //
@@ -29,6 +29,8 @@ import java.awt.*;
 import java.awt.event.*;
 import java.io.*;
 import java.net.*;
+import java.util.*;
+import netscape.javascript.JSObject;
 
 public class VncViewer extends java.applet.Applet
   implements java.lang.Runnable, WindowListener {
@@ -61,7 +63,9 @@ public class VncViewer extends java.applet.Applet
   ScrollPane desktopScrollPane;
   GridBagLayout gridbag;
   ButtonPanel buttonPanel;
+  Label connStatusLabel;
   AuthPanel authenticator;
+  AuthUnixLoginPanel authenticatorUnixLogin;
   VncCanvas vc;
   OptionsFrame options;
   ClipboardFrame clipboard;
@@ -79,8 +83,6 @@ public class VncViewer extends java.applet.Applet
   String socketFactory;
   String host;
   int port;
-  String passwordParam;
-  String encPasswordParam;
   boolean showControls;
   boolean offerRelogin;
   boolean showOfflineDesktop;
@@ -88,6 +90,25 @@ public class VncViewer extends java.applet.Applet
   int deferCursorUpdates;
   int deferUpdateRequests;
 
+  // A variable to be passed on to detect if this applet is signed or not
+  //    -- Sundar
+  boolean isSigned;
+
+
+  boolean disableSSL;
+  boolean setServerSize;
+  boolean GET;
+  String CONNECT;
+  String urlPrefix;
+  String httpsPort;
+  boolean forceProxy;
+  boolean trustAllVncCerts;
+
+  // Reference to this applet for inter-applet communication.
+  public static java.applet.Applet refApplet;
+
+  // For JavaScript interaction.
+  JSObject jso = null;
 
   //
   // init()
@@ -95,8 +116,23 @@ public class VncViewer extends java.applet.Applet
 
   public void init() {
 
+    if (inAnApplet) {
+      if (jso == null) {
+        try {
+          jso = JSObject.getWindow(this);
+        } catch(Exception e) {
+          String msg = e.getMessage();
+          System.out.println("Can't get JSObject: " + msg);
+        }
+      }
+
+      javascript("cancelTimeout()");
+    }
+
     readParameters();
 
+    refApplet = this;
+
     if (inSeparateFrame) {
       vncFrame = new Frame("TightVNC");
       if (!inAnApplet) {
@@ -111,7 +147,8 @@ public class VncViewer extends java.applet.Applet
 
     options = new OptionsFrame(this);
     clipboard = new ClipboardFrame(this);
-    authenticator = new AuthPanel();
+    authenticator = new AuthPanel(this);
+    authenticatorUnixLogin = new AuthUnixLoginPanel();
     if (RecordingFrame.checkSecurity())
       rec = new RecordingFrame(this);
 
@@ -152,7 +189,6 @@ public class VncViewer extends java.applet.Applet
 
     try {
       connectAndAuthenticate();
-
       doProtocolInitialisation();
 
       vc = new VncCanvas(this);
@@ -192,7 +228,8 @@ public class VncViewer extends java.applet.Applet
       if (showControls)
 	buttonPanel.enableButtons();
 
-      vc.processNormalProtocol();
+      moveFocusToDesktop();
+      processNormalProtocol();
 
     } catch (NoRouteToHostException e) {
       fatalError("Network error: no route to server: " + host, e);
@@ -243,54 +280,182 @@ public class VncViewer extends java.applet.Applet
   }
 
 
+  //
+  // Process RFB socket messages.
+  // If the rfbThread is being stopped, ignore any exceptions,
+  // otherwise rethrow the exception so it can be handled.
+  //
+ 
+  void processNormalProtocol() throws Exception {
+    while(true) {
+      try {
+        vc.processNormalProtocol();
+      } catch (ArrayIndexOutOfBoundsException e) {
+        System.out.println("Out of bounds exception.");
+      } catch (Exception e) {
+        if (rfbThread == null) {
+	  System.out.println("Ignoring RFB socket exceptions" +
+			     " because applet is stopping");
+          return;
+        } else {
+	  throw e;
+        }
+      }
+    }
+  }
+
+
   //
   // Connect to the RFB server and authenticate the user.
   //
 
-  void connectAndAuthenticate() throws Exception {
-
-    // If "ENCPASSWORD" parameter is set, decrypt the password into
-    // the passwordParam string.
-
-    if (encPasswordParam != null) {
-      // ENCPASSWORD is hexascii-encoded. Decode.
-      byte[] pw = {0, 0, 0, 0, 0, 0, 0, 0};
-      int len = encPasswordParam.length() / 2;
-      if (len > 8)
-	len = 8;
-      for (int i = 0; i < len; i++) {
-	String hex = encPasswordParam.substring(i*2, i*2+2);
-	Integer x = new Integer(Integer.parseInt(hex, 16));
-	pw[i] = x.byteValue();
-      }
-
-      // Decrypt the password.
-      byte[] key = {23, 82, 107, 6, 35, 78, 88, 7};
-      DesCipher des = new DesCipher(key);
-      des.decrypt(pw, 0, pw, 0);
-      passwordParam = new String(pw);
+  void connectAndAuthenticate() throws Exception
+  {
+    showConnectionStatus("Initializing...");
+    if (inSeparateFrame) {
+      vncFrame.pack();
+      vncFrame.show();
+    } else {
+      validate();
     }
 
-    // If a password parameter ("PASSWORD" or "ENCPASSWORD") is set,
-    // don't ask user a password, get one from passwordParam instead.
-    // Authentication failures would be fatal.
+    while (!tryAuthenticate()) {
+      authenticator.retry();
+      authenticatorUnixLogin.retry();
+    }
+  }
 
-    if (passwordParam != null) {
-      if (inSeparateFrame) {
-	vncFrame.pack();
-	vncFrame.show();
-      } else {
-	validate();
+
+  //
+  // Try to connect and authenticate.
+  //
+
+  boolean tryAuthenticate() throws Exception
+  {
+    showConnectionStatus("Connecting to " + host + ", port " + port + "...");
+
+    rfb = new RfbProto(host, port, this);
+    showConnectionStatus("Connected to server");
+
+    rfb.readVersionMsg();
+    showConnectionStatus("RFB server supports protocol version " +
+			 rfb.serverMajor + "." + rfb.serverMinor);
+
+    rfb.writeVersionMsg();
+    showConnectionStatus("Using RFB protocol version " +
+			 rfb.clientMajor + "." + rfb.clientMinor);
+
+    int secType = rfb.negotiateSecurity();
+    int authType = 0;
+
+    // FIXME: Map security types to authentication schemes in RfbProto.
+    switch (secType) {
+    case RfbProto.SecTypeNone:
+      authType = RfbProto.AuthNone;
+      break;
+    case RfbProto.SecTypeVncAuth:
+      authType = RfbProto.AuthVNC;
+      break;
+    case RfbProto.SecTypeTight:
+      showConnectionStatus("Enabling TightVNC protocol extensions");
+      rfb.initCapabilities();
+      rfb.setupTunneling();
+      authType = rfb.negotiateAuthenticationTight();
+      break;
+    default:
+      throw new Exception("Unknown security type " + secType);
+    }
+
+    boolean success = false;
+
+    switch (authType) {
+    case RfbProto.AuthNone:
+      showConnectionStatus("No authentication needed");
+      success = true;
+      break;
+    case RfbProto.AuthVNC:
+      showConnectionStatus("Performing standard VNC authentication");
+      if (authenticator.isInteractionNecessary()) {
+	showAuthPanel(authenticator);
+	authenticator.moveFocusToDefaultField();
       }
-      if (!tryAuthenticate(passwordParam)) {
-	throw new Exception("VNC authentication failed");
+      success = authenticator.tryAuthenticate(rfb);
+      if (authenticator.isInteractionNecessary()) {
+	vncContainer.remove(authenticator);
+      } else {
+	// Don't retry non-interactive authentication.
+	if (!success)
+	  throw new Exception("VNC authentication failed");
+      }
+      break;
+    case RfbProto.AuthUnixLogin:
+      showConnectionStatus("Performing Unix login-style authentication");
+      showAuthPanel(authenticatorUnixLogin);
+      authenticatorUnixLogin.moveFocusToDefaultField();
+      success = authenticatorUnixLogin.tryAuthenticate(rfb);
+      vncContainer.remove(authenticatorUnixLogin);
+      break;
+    default:
+      throw new Exception("Unknown authentication scheme " + authType);
+    }
+
+    if (!success)
+      rfb.close();
+
+    return success;
+  }
+
+
+  //
+  // Show a message describing the connection status.
+  // To hide the connection status label, use (msg == null).
+  //
+
+  void showConnectionStatus(String msg)
+  {
+    if (msg == null) {
+      if (vncContainer.isAncestorOf(connStatusLabel)) {
+	vncContainer.remove(connStatusLabel);
       }
       return;
     }
 
-    // There is no "PASSWORD" or "ENCPASSWORD" parameters -- ask user
-    // for a password, try to authenticate, retry on authentication
-    // failures.
+    System.out.println(msg);
+
+    if (connStatusLabel == null) {
+      connStatusLabel = new Label("Status: " + msg);
+      connStatusLabel.setFont(new Font("Helvetica", Font.PLAIN, 12));
+    } else {
+      connStatusLabel.setText("Status: " + msg);
+    }
+
+    if (!vncContainer.isAncestorOf(connStatusLabel)) {
+      GridBagConstraints gbc = new GridBagConstraints();
+      gbc.gridwidth = GridBagConstraints.REMAINDER;
+      gbc.fill = GridBagConstraints.HORIZONTAL;
+      gbc.anchor = GridBagConstraints.NORTHWEST;
+      gbc.weightx = 1.0;
+      gbc.weighty = 1.0;
+      gbc.insets = new Insets(20, 30, 20, 30);
+      gridbag.setConstraints(connStatusLabel, gbc);
+      vncContainer.add(connStatusLabel);
+    }
+
+    if (inSeparateFrame) {
+      vncFrame.pack();
+    } else {
+      validate();
+    }
+  }
+
+
+  //
+  // Show an authentication panel.
+  //
+
+  void showAuthPanel(Panel authPanel)
+  {
+    showConnectionStatus(null);
 
     GridBagConstraints gbc = new GridBagConstraints();
     gbc.gridwidth = GridBagConstraints.REMAINDER;
@@ -299,111 +464,14 @@ public class VncViewer extends java.applet.Applet
     gbc.weighty = 1.0;
     gbc.ipadx = 100;
     gbc.ipady = 50;
-    gridbag.setConstraints(authenticator, gbc);
-    vncContainer.add(authenticator);
+    gridbag.setConstraints(authPanel, gbc);
+    vncContainer.add(authPanel);
 
     if (inSeparateFrame) {
       vncFrame.pack();
-      vncFrame.show();
     } else {
       validate();
-      // FIXME: here moveFocusToPasswordField() does not always work
-      // under Netscape 4.7x/Java 1.1.5/Linux. It seems like this call
-      // is being executed before the password field of the
-      // authenticator is fully drawn and activated, therefore
-      // requestFocus() does not work. Currently, I don't know how to
-      // solve this problem.
-      //   -- const
-      authenticator.moveFocusToPasswordField();
     }
-
-    while (true) {
-      // Wait for user entering a password.
-      synchronized(authenticator) {
-	try {
-	  authenticator.wait();
-	} catch (InterruptedException e) {
-	}
-      }
-
-      // Try to authenticate with a given password.
-      if (tryAuthenticate(authenticator.password.getText()))
-	break;
-
-      // Retry on authentication failure.
-      authenticator.retry();
-    }
-
-    vncContainer.remove(authenticator);
-  }
-
-
-  //
-  // Try to authenticate with a given password.
-  //
-
-  boolean tryAuthenticate(String pw) throws Exception {
-
-    rfb = new RfbProto(host, port, this);
-
-    rfb.readVersionMsg();
-
-    System.out.println("RFB server supports protocol version " +
-		       rfb.serverMajor + "." + rfb.serverMinor);
-
-    rfb.writeVersionMsg();
-
-    int authScheme = rfb.readAuthScheme();
-
-    switch (authScheme) {
-
-    case RfbProto.NoAuth:
-      System.out.println("No authentication needed");
-      return true;
-
-    case RfbProto.VncAuth:
-      byte[] challenge = new byte[16];
-      rfb.is.readFully(challenge);
-
-      if (pw.length() > 8)
-	pw = pw.substring(0, 8); // Truncate to 8 chars
-
-      // vncEncryptBytes in the UNIX libvncauth truncates password
-      // after the first zero byte. We do to.
-      int firstZero = pw.indexOf(0);
-      if (firstZero != -1)
-	pw = pw.substring(0, firstZero);
-
-      byte[] key = {0, 0, 0, 0, 0, 0, 0, 0};
-      System.arraycopy(pw.getBytes(), 0, key, 0, pw.length());
-
-      DesCipher des = new DesCipher(key);
-
-      des.encrypt(challenge, 0, challenge, 0);
-      des.encrypt(challenge, 8, challenge, 8);
-
-      rfb.os.write(challenge);
-
-      int authResult = rfb.is.readInt();
-
-      switch (authResult) {
-      case RfbProto.VncAuthOK:
-	System.out.println("VNC authentication succeeded");
-	return true;
-      case RfbProto.VncAuthFailed:
-	System.out.println("VNC authentication failed");
-	break;
-      case RfbProto.VncAuthTooMany:
-	throw new Exception("VNC authentication failed - too many tries");
-      default:
-	throw new Exception("Unknown VNC authentication result " + authResult);
-      }
-      break;
-
-    default:
-      throw new Exception("Unknown VNC authentication scheme " + authScheme);
-    }
-    return false;
   }
 
 
@@ -411,17 +479,24 @@ public class VncViewer extends java.applet.Applet
   // Do the rest of the protocol initialisation.
   //
 
-  void doProtocolInitialisation() throws IOException {
-
+  void doProtocolInitialisation() throws IOException
+  {
     rfb.writeClientInit();
-
     rfb.readServerInit();
 
+    if (setServerSize) {
+      requestResize(getSize().width, getSize().height);
+      rfb.framebufferWidth = getSize().width;
+      rfb.framebufferHeight = getSize().height;
+    }
+
     System.out.println("Desktop name is " + rfb.desktopName);
     System.out.println("Desktop size is " + rfb.framebufferWidth + " x " +
 		       rfb.framebufferHeight);
 
     setEncodings();
+
+    showConnectionStatus(null);
   }
 
 
@@ -429,16 +504,113 @@ public class VncViewer extends java.applet.Applet
   // Send current encoding list to the RFB server.
   //
 
-  void setEncodings() {
-    try {
-      if (rfb != null && rfb.inNormalProtocol) {
-	rfb.writeSetEncodings(options.encodings, options.nEncodings);
-	if (vc != null) {
-	  vc.softCursorFree();
-	}
+  int[] encodingsSaved;
+  int nEncodingsSaved;
+
+  void setEncodings()        { setEncodings(false); }
+  void autoSelectEncodings() { setEncodings(true); }
+
+  void setEncodings(boolean autoSelectOnly) {
+    if (options == null || rfb == null || !rfb.inNormalProtocol)
+      return;
+
+    int preferredEncoding = options.preferredEncoding;
+    if (preferredEncoding == -1) {
+      long kbitsPerSecond = rfb.kbitsPerSecond();
+      if (nEncodingsSaved < 1) {
+        // Choose Tight encoding for the very first update.
+        System.out.println("Using Tight encoding");
+        preferredEncoding = RfbProto.EncodingTight;
+      } else if (kbitsPerSecond > 2000 &&
+                 encodingsSaved[0] != RfbProto.EncodingHextile) {
+        // Switch to Hextile if the connection speed is above 2Mbps.
+        System.out.println("Throughput " + kbitsPerSecond +
+                           " kbit/s - changing to Hextile encoding");
+        preferredEncoding = RfbProto.EncodingHextile;
+      } else if (kbitsPerSecond < 1000 &&
+                 encodingsSaved[0] != RfbProto.EncodingTight) {
+        // Switch to Tight if the connection speed is below 1Mbps.
+        System.out.println("Throughput " + kbitsPerSecond +
+                           " kbit/s - changing to Tight encoding");
+        preferredEncoding = RfbProto.EncodingTight;
+      } else {
+        // Don't change the encoder.
+        if (autoSelectOnly)
+          return;
+        preferredEncoding = encodingsSaved[0];
       }
-    } catch (Exception e) {
-      e.printStackTrace();
+    } else {
+      // Auto encoder selection is not enabled.
+      if (autoSelectOnly)
+        return;
+    }
+
+    int[] encodings = new int[20];
+    int nEncodings = 0;
+
+    encodings[nEncodings++] = preferredEncoding;
+    if (options.useCopyRect) {
+      encodings[nEncodings++] = RfbProto.EncodingCopyRect;
+    }
+
+    if (preferredEncoding != RfbProto.EncodingHextile) {
+      encodings[nEncodings++] = RfbProto.EncodingHextile;
+    }
+    if (preferredEncoding != RfbProto.EncodingTight) {
+      encodings[nEncodings++] = RfbProto.EncodingTight;
+    }
+    if (preferredEncoding != RfbProto.EncodingZlib) {
+      encodings[nEncodings++] = RfbProto.EncodingZlib;
+    }
+    if (preferredEncoding != RfbProto.EncodingCoRRE) {
+      encodings[nEncodings++] = RfbProto.EncodingCoRRE;
+    }
+    if (preferredEncoding != RfbProto.EncodingRRE) {
+      encodings[nEncodings++] = RfbProto.EncodingRRE;
+    }
+
+    if (options.compressLevel >= 0 && options.compressLevel <= 9) {
+      encodings[nEncodings++] =
+        RfbProto.EncodingCompressLevel0 + options.compressLevel;
+    }
+    if (options.jpegQuality >= 0 && options.jpegQuality <= 9) {
+      encodings[nEncodings++] =
+        RfbProto.EncodingQualityLevel0 + options.jpegQuality;
+    }
+
+    if (options.requestCursorUpdates) {
+      encodings[nEncodings++] = RfbProto.EncodingXCursor;
+      encodings[nEncodings++] = RfbProto.EncodingRichCursor;
+      if (!options.ignoreCursorUpdates)
+	encodings[nEncodings++] = RfbProto.EncodingPointerPos;
+    }
+
+    encodings[nEncodings++] = RfbProto.EncodingLastRect;
+    encodings[nEncodings++] = RfbProto.EncodingNewFBSize;
+
+    boolean encodingsWereChanged = false;
+    if (nEncodings != nEncodingsSaved) {
+      encodingsWereChanged = true;
+    } else {
+      for (int i = 0; i < nEncodings; i++) {
+        if (encodings[i] != encodingsSaved[i]) {
+          encodingsWereChanged = true;
+          break;
+        }
+      }
+    }
+
+    if (encodingsWereChanged) {
+      try {
+        rfb.writeSetEncodings(encodings, nEncodings);
+        if (vc != null) {
+          vc.softCursorFree();
+        }
+      } catch (Exception e) {
+        e.printStackTrace();
+      }
+      encodingsSaved = encodings;
+      nEncodingsSaved = nEncodings;
     }
   }
 
@@ -568,10 +740,6 @@ public class VncViewer extends java.applet.Applet
 	inSeparateFrame = true;
     }
 
-    encPasswordParam = readParameter("ENCPASSWORD", false);
-    if (encPasswordParam == null)
-      passwordParam = readParameter("PASSWORD", false);
-
     // "Show Controls" set to "No" disables button panel.
     showControls = true;
     str = readParameter("Show Controls", false);
@@ -598,6 +766,67 @@ public class VncViewer extends java.applet.Applet
 
     // SocketFactory.
     socketFactory = readParameter("SocketFactory", false);
+
+    // SSL
+    disableSSL = true;
+    str = readParameter("DisableSSL", false);
+    if (str != null && str.equalsIgnoreCase("No"))
+      disableSSL = false;
+
+    httpsPort = readParameter("httpsPort", false);
+
+    // Extra GET, CONNECT string:
+    CONNECT = readParameter("CONNECT", false);
+    if (CONNECT != null) {
+	CONNECT = CONNECT.replaceAll(" ", ":");
+    }
+
+    GET = false;
+    str = readParameter("GET", false);
+    if (str != null && str.equalsIgnoreCase("Yes")) {
+      GET = true;
+    }
+    if (str != null && str.equalsIgnoreCase("1")) {
+      GET = true;
+    }
+
+    urlPrefix = readParameter("urlPrefix", false);
+    if (urlPrefix != null) {
+	urlPrefix = urlPrefix.replaceAll("%2F", "/");
+	urlPrefix = urlPrefix.replaceAll("%2f", "/");
+	urlPrefix = urlPrefix.replaceAll("_2F_", "/");
+	if (urlPrefix.indexOf("/") != 0) {
+		urlPrefix = "/" + urlPrefix;
+	}
+    } else {
+        urlPrefix = "";
+    }
+    System.out.println("urlPrefix: " + urlPrefix);
+
+    forceProxy = false;
+    str = readParameter("forceProxy", false);
+    if (str != null && str.equalsIgnoreCase("Yes")) {
+      forceProxy = true;
+    }
+    trustAllVncCerts = false;
+    str = readParameter("trustAllVncCerts", false);
+    if (str != null && str.equalsIgnoreCase("Yes")) {
+      trustAllVncCerts = true;
+    }
+    // Reading the parameter that indicates if the Applet is signed or not
+    // --Sundar
+    isSigned = false;
+    str = readParameter("signed", false);
+    if (str != null && str.equalsIgnoreCase("Yes")) {
+      isSigned = true;
+    }
+
+    // SetServerSize
+    setServerSize = false;
+    str = readParameter("SetServerSize", false);
+    if (str != null && str.equalsIgnoreCase("Yes")) {
+      setServerSize = true;
+    }
   }
 
   public String readParameter(String name, boolean required) {
@@ -645,10 +874,14 @@ public class VncViewer extends java.applet.Applet
 
   void moveFocusToDesktop() {
     if (vncContainer != null) {
+      vncContainer.setFocusCycleRoot(true);
+      vncContainer.setFocusTraversalKeysEnabled(false);
       if (vc != null && vncContainer.isAncestorOf(vc)) {
 	vc.requestFocus();
       } else if (vncContainer.isAncestorOf(authenticator)) {
-	authenticator.moveFocusToPasswordField();
+	authenticator.moveFocusToDefaultField();
+      } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
+	authenticatorUnixLogin.moveFocusToDefaultField();
       }
     }
   }
@@ -676,6 +909,7 @@ public class VncViewer extends java.applet.Applet
 
   //
   // fatalError() - print out a fatal error message.
+  // FIXME: Do we really need two versions of the fatalError() method?
   //
 
   synchronized public void fatalError(String str) {
@@ -692,18 +926,18 @@ public class VncViewer extends java.applet.Applet
 
   synchronized public void fatalError(String str, Exception e) {
  
-    if (rfb != null) {
+    if (rfb != null && rfb.closed()) {
       // Not necessary to show error message if the error was caused
       // by I/O problems after the rfb.close() method call.
-      if (rfb.closed()) {
-	System.out.println("RFB thread finished");
-	return;
-      }
-      rfb.close();
+      System.out.println("RFB thread finished");
+      return;
     }
 
-    e.printStackTrace();
     System.out.println(str);
+    e.printStackTrace();
+
+    if (rfb != null)
+      rfb.close();
 
     if (inAnApplet) {
       showMessage(str);
@@ -719,26 +953,25 @@ public class VncViewer extends java.applet.Applet
   void showMessage(String msg) {
     vncContainer.removeAll();
 
-    Label errLabel = new Label(msg, Label.CENTER);
-    errLabel.setFont(new Font("Helvetica", Font.PLAIN, 12));
+    // Make sure the browser is displaying the applet.
+    forceSize();
 
     if (offerRelogin) {
-
-      Panel gridPanel = new Panel(new GridLayout(0, 1));
-      Panel outerPanel = new Panel(new FlowLayout(FlowLayout.LEFT));
-      outerPanel.add(gridPanel);
-      vncContainer.setLayout(new FlowLayout(FlowLayout.LEFT, 30, 16));
-      vncContainer.add(outerPanel);
-      Panel textPanel = new Panel(new FlowLayout(FlowLayout.CENTER));
-      textPanel.add(errLabel);
-      gridPanel.add(textPanel);
-      gridPanel.add(new ReloginPanel(this));
+      ReloginPanel reloginPanel = new ReloginPanel(this, msg);
+      vncContainer.add(reloginPanel);
+      GridBagConstraints gbc = new GridBagConstraints();
+      gbc.gridwidth = GridBagConstraints.REMAINDER;
+      gbc.fill = GridBagConstraints.BOTH;
+      gbc.anchor = GridBagConstraints.NORTHWEST;
+      gbc.weightx = 1.0;
+      gbc.weighty = 1.0;
+      gbc.insets = new Insets(8, 8, 8, 8);
+      gridbag.setConstraints(reloginPanel, gbc);
 
     } else {
-
-      vncContainer.setLayout(new FlowLayout(FlowLayout.LEFT, 30, 30));
+      Label errLabel = new Label(msg, Label.CENTER);
+      errLabel.setFont(new Font("Helvetica", Font.PLAIN, 12));
       vncContainer.add(errLabel);
-
     }
 
     if (inSeparateFrame) {
@@ -748,12 +981,24 @@ public class VncViewer extends java.applet.Applet
     }
   }
 
+  //
+  // Stop the applet.
+  // Main applet thread will terminate on first exception
+  // after seeing that rfbThread has been set to null.
+  //
+
+  public void stop() {
+    System.out.println("Stopping applet");
+    rfbThread = null;
+  }
+
   //
   // This method is called before the applet is destroyed.
   //
 
   public void destroy() {
     System.out.println("Destroying applet");
+
     vncContainer.removeAll();
     options.dispose();
     clipboard.dispose();
@@ -765,6 +1010,13 @@ public class VncViewer extends java.applet.Applet
       vncFrame.dispose();
   }
 
+  //
+  // Start/stop receiving mouse events.
+  //
+
+  public void enableInput(boolean enable) {
+    vc.enableInput(enable);
+  }
 
   //
   // Close application properly on window close event.
@@ -775,7 +1027,8 @@ public class VncViewer extends java.applet.Applet
     if (rfb != null)
       disconnect();
 
-    vncFrame.dispose();
+    vncContainer.hide();
+
     if (!inAnApplet) {
       System.exit(0);
     }
@@ -786,8 +1039,11 @@ public class VncViewer extends java.applet.Applet
   //
 
   public void windowActivated(WindowEvent evt) {
-    if (vncFrame.isAncestorOf(authenticator))
-      authenticator.moveFocusToPasswordField();
+    if (vncFrame.isAncestorOf(authenticator)) {
+      authenticator.moveFocusToDefaultField();
+    } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
+      authenticatorUnixLogin.moveFocusToDefaultField();
+    }
   }
 
   //
@@ -799,4 +1055,169 @@ public class VncViewer extends java.applet.Applet
   public void windowClosed(WindowEvent evt) {}
   public void windowIconified(WindowEvent evt) {}
   public void windowDeiconified(WindowEvent evt) {}
+
+  //================================================================
+  // The methods below were added for HUB application resize support.
+  //================================================================
+
+  // Javascript calls this to refresh the screen.
+  public void refresh() {
+    if (rfb == null)
+      return;
+    try {
+      rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
+                                        rfb.framebufferHeight, false);
+    }
+    catch(IOException e) {
+    }
+  }
+
+  // Javascript calls this to request a resize from the VNC server.
+  public void requestResize(int w, int h) {
+    if (rfb == null)
+      return;
+
+    if ((w > 4096) || (w < 0)) {
+      System.out.println("ERROR in requestResize: Invalid width " +
+                         Integer.toString(w));
+    }
+    if ((h > 4096) || (h < 0)) {
+      System.out.println("ERROR in requestResize: Invalid height " +
+                         Integer.toString(w));
+    }
+
+    try {
+      rfb.writeRequestFBSize(w,h);
+    }
+    catch(IOException e) {
+    }
+  }
+
+  // Called internally to force the browser to resize the Applet.
+  public void forceSize() {
+    int w;
+    int h;
+
+    if (rfb == null) {
+        w = getSize().width;
+        h = getSize().height;
+    } else {
+        w = rfb.framebufferWidth;
+        h = rfb.framebufferHeight;
+    }
+
+    if (inSeparateFrame) {
+        if (vc != null) /* *njk* *hack* check needed when not run as applet */
+            vc.setSize(w, h);
+        vncFrame.pack();
+        refresh();
+        return;
+    }
+
+    setSize(w, h);
+
+    String ws = Integer.toString(w);
+    String hs = Integer.toString(h);
+    System.out.println("forceSize(" + ws + "," + hs + ")");
+    javascript("forceSize(" + ws + "," + hs + ")");
+
+    setSize(w, h);
+
+    refresh();
+  }
+
+  // Called internally to invoke a client action.
+  public void clientAction(String action) {
+    javascript("clientAction('" + action + "')");
+  }
+
+    // Helper function to execute javascript in the browser.
+    public void javascript(String s) {
+	if (!inAnApplet) {
+            return;
+        }
+        if (jso == null) {
+            System.out.println("JSObject is NULL");
+            return;
+        }
+        try {
+          jso.eval(s);
+          System.out.println("Executed: " + s);
+        }
+        catch(Exception e) {
+          System.out.println("javascript: Can't call: " + s);
+        }
+    }
+
+    // Called externally to switch between applet view and separate frame.
+    public void popout() {
+
+        vncContainer.removeAll();
+
+        if (inSeparateFrame) {
+            inSeparateFrame = false;
+
+            vncFrame.dispose();
+            vncFrame = null;
+            vncContainer = this;
+        } else {
+            inSeparateFrame = true;
+
+            vncFrame = new Frame("TightVNC");
+            vncFrame.setTitle(rfb.desktopName);
+            vncFrame.setResizable(true);
+            vncFrame.addWindowListener(this);
+            vncContainer = vncFrame;
+        }
+
+        gridbag = new GridBagLayout();
+        vncContainer.setLayout(gridbag);
+
+        GridBagConstraints gbc = new GridBagConstraints();
+        gbc.gridwidth = GridBagConstraints.REMAINDER;
+        gbc.anchor = GridBagConstraints.NORTHWEST;
+
+        if (showControls) {
+            buttonPanel = new ButtonPanel(this);
+            gridbag.setConstraints(buttonPanel, gbc);
+            vncContainer.add(buttonPanel);
+        }
+
+        gbc.weightx = 1.0;
+        gbc.weighty = 1.0;
+        gbc.fill = GridBagConstraints.BOTH;
+        gridbag.setConstraints(vc, gbc);
+        vncContainer.add(vc);
+
+        if (inSeparateFrame) {
+            vncFrame.pack();
+            vncFrame.show();
+
+            vc.addComponentListener(new ComponentAdapter()
+                {
+                    Timer updateDelay = null;
+                    public void componentResized(ComponentEvent e)
+                    {
+                        final Component c = e.getComponent();
+                        if (updateDelay != null)
+                            updateDelay.cancel();
+                        updateDelay = new Timer();
+                        //System.out.println("Scheduled resize.");
+                        updateDelay.schedule(new TimerTask()
+                            {
+                                public void run() {
+                                    //System.out.println("RESIZE");
+                                    requestResize(c.getWidth(), c.getHeight());
+                                    updateDelay = null;
+                                }
+                            }, 1000);
+                    }
+                });
+        } else {
+            validate();
+            forceSize();
+        }
+
+    }
 }
+
